from unittest import mock

import pytest

from ape.api import TransactionAPI
from ape.exceptions import PluginVersionError
from ape.logging import LogLevel
from ape.managers.plugins import _get_unimplemented_methods_warning
from ape.plugins._utils import CORE_PLUGINS as CORE_PLUGINS_LIST
from ape.plugins._utils import (
    ApePluginsRepr,
    ModifyPluginResultHandler,
    PluginGroup,
    PluginMetadata,
    PluginMetadataList,
    PluginType,
    _filter_plugins_from_dists,
    ape_version,
)

CORE_PLUGINS = ("run",)
AVAILABLE_PLUGINS = ("available", "installed")
INSTALLED_PLUGINS = ("installed", "thirdparty")
THIRD_PARTY = ("thirdparty",)


mark_specifiers_less_than_ape = pytest.mark.parametrize(
    "specifier",
    (f"<{ape_version[0]}", f">0.1,<{ape_version[0]}", f"==0.{int(ape_version[2]) - 1}"),
)
parametrize_pip_cmd = pytest.mark.parametrize(
    "pip_command", [["python", "-m", "pip"], ["uv", "pip"]]
)


@pytest.fixture(autouse=True)
def get_dists_patch(mocker):
    return mocker.patch("ape.plugins._utils._get_distributions")


@pytest.fixture
def make_dist(mocker):
    def fn(name, version):
        mock_dist = mocker.MagicMock()
        mock_dist.name = name
        mock_dist.version = version
        return mock_dist

    return fn


@pytest.fixture(autouse=True)
def mock_installed_packages(make_dist, get_dists_patch):
    def fn(version: str):
        get_dists_patch.return_value = (
            make_dist("FOOFOO", "1.1.1"),
            make_dist(f"ape-{INSTALLED_PLUGINS[0]}", version),
            make_dist("aiohttp", "3.8.5"),
            make_dist(f"ape-{THIRD_PARTY[0]}", version),
        )
        return get_dists_patch

    return fn


@pytest.fixture(autouse=True)
def plugin_test_env(mocker, mock_installed_packages):
    root = "ape.plugins._utils"

    # Prevent calling out to GitHub
    gh_mock = mocker.patch(f"{root}._get_available_plugins")
    gh_mock.return_value = {f"ape_{x}" for x in AVAILABLE_PLUGINS}

    mock_installed_packages(ape_version.base)


@pytest.fixture
def package_names() -> set[str]:
    return {
        f"ape-{x}" for x in [*CORE_PLUGINS, *AVAILABLE_PLUGINS, *INSTALLED_PLUGINS, *THIRD_PARTY]
    }


@pytest.fixture
def plugin_metadata(package_names, plugin_test_env) -> PluginMetadataList:
    names = {x for x in package_names}
    names.remove("ape-installed")
    names.add(f"ape-installed==0.{ape_version.minor}.0")
    names.remove("ape-thirdparty")
    names.add(f"ape-thirdparty==0.{ape_version.minor}.0")
    return PluginMetadataList.from_package_names(names, trusted_list=("installed",))


@pytest.fixture
def mock_python_path(tmp_path):
    """Create a temporary path to simulate a Python interpreter location."""
    python_path = tmp_path / "python"
    python_path.touch()
    return f"{python_path}"


@pytest.fixture
def mock_nonexistent_path():
    """Return a path that definitely doesn't exist."""
    return "/path/that/does/not/exist/python"


class TestPluginMetadataList:
    def test_from_package_names(self, plugin_metadata):
        actual = plugin_metadata
        assert actual.core.plugin_names == list(CORE_PLUGINS)
        assert actual.third_party.plugin_names == list(THIRD_PARTY)
        assert actual.installed.plugin_names == [INSTALLED_PLUGINS[0]]  # Not 3rd party
        assert actual.available.plugin_names == [AVAILABLE_PLUGINS[0]]  # Not installed

    def test_all_plugins(self, plugin_metadata, package_names):
        actual = {f"ape-{x.name}" for x in plugin_metadata.all_plugins}
        assert actual == package_names


class TestPluginMetadata:
    @pytest.mark.parametrize(
        "name", ("ape-foo-bar", "ape-foo-bar", "ape_foo_bar", "foo-bar", "foo_bar")
    )
    def test_names(self, name):
        metadata = PluginMetadata(name=name)
        assert metadata.name == "foo-bar"
        assert metadata.package_name == "ape-foo-bar"
        assert metadata.module_name == "ape_foo_bar"

    def test_model_validator_when_version_included_with_name(self):
        # This allows parsing requirements files easier
        metadata = PluginMetadata(name=f"ape-foo-bar==0.{ape_version.minor}.0")
        assert metadata.name == "foo-bar"
        assert metadata.version == f"==0.{ape_version.minor}.0"

    @pytest.mark.parametrize(
        "version",
        (f"0.{ape_version.minor}.0", f"v0.{ape_version.minor}.0", f"0.{ape_version.minor}.0a123"),
    )
    def test_version(self, version):
        metadata = PluginMetadata(name="foo", version=version)
        assert metadata.version == version

    def test_install_str_without_version(self):
        metadata = PluginMetadata(name="foo-bar")
        actual = metadata.install_str
        expected_version = f">=0.{ape_version.minor},<0.{ape_version.minor + 1}"
        assert actual == f"ape-foo-bar{expected_version}"

    def test_install_str_with_version(self):
        metadata = PluginMetadata(name="foo-bar", version=f"0.{ape_version.minor}.0")
        actual = metadata.install_str
        assert actual == f"ape-foo-bar==0.{ape_version.minor}.0"

    def test_install_str_with_complex_constraint(self):
        metadata = PluginMetadata(
            name="foo", version=f">=0.{ape_version.minor}.0,<0.{ape_version.minor + 1}.0"
        )
        actual = metadata.install_str
        assert actual == f"ape-foo>=0.{ape_version.minor}.0,<0.{ape_version.minor + 1}.0"

    def test_install_str_with_complex_constraint_in_name(self):
        metadata = PluginMetadata(name=f"foo>=0.{ape_version.minor}.0,<0.{ape_version.minor + 1}.0")
        actual = metadata.install_str
        assert actual == f"ape-foo>=0.{ape_version.minor}.0,<0.{ape_version.minor + 1}.0"

    def test_install_str_when_using_git_remote(self):
        url = "git+https://example.com/myorg/ape-foo.git@branch_name"
        metadata = PluginMetadata(name="foo", version=url)
        assert metadata.name == "foo"
        actual = metadata.install_str
        assert actual == url

    def test_install_str_remote_in_name(self):
        url = "git+https://example.com/myorg/ape-foo.git@branch_name"
        metadata = PluginMetadata(name=f"foo@{url}")
        assert metadata.name == "foo"
        actual = metadata.install_str
        assert actual == url

    def test_install_str_only_remote(self):
        url = "git+https://example.com/myorg/ape-foo.git@branch_name"
        metadata = PluginMetadata(name=url)
        actual = metadata.install_str
        assert actual == url
        assert metadata.name == "foo"

    def test_is_available(self):
        metadata = PluginMetadata(name=list(AVAILABLE_PLUGINS)[0])
        assert metadata.is_available
        metadata = PluginMetadata(name="foobar")
        assert not metadata.is_available

    @parametrize_pip_cmd
    def test_prepare_install(self, pip_command):
        metadata = PluginMetadata(name=list(AVAILABLE_PLUGINS)[0], pip_command=pip_command)
        actual = metadata._prepare_install(skip_confirmation=True)
        assert actual is not None
        arguments = actual.get("args", [])
        shared = [
            "pip",
            "install",
            f"ape-available>=0.{ape_version.minor},<0.{ape_version.minor + 1}",
            "--quiet",
        ]
        if arguments[0] == "uv":
            expected = ["uv", *shared]
            assert arguments == expected
        else:
            expected = ["-m", *shared]
            assert "python" in arguments[0]
            assert arguments[1:] == expected

    @parametrize_pip_cmd
    def test_prepare_install_upgrade(self, pip_command):
        metadata = PluginMetadata(name=list(AVAILABLE_PLUGINS)[0], pip_command=pip_command)
        actual = metadata._prepare_install(upgrade=True, skip_confirmation=True)
        assert actual is not None
        arguments = actual.get("args", [])
        shared = [
            "pip",
            "install",
            "--upgrade",
            f"ape-available>=0.{ape_version.minor},<0.{ape_version.minor + 1}",
            "--quiet",
        ]

        if pip_command[0].startswith("uv"):
            expected = ["uv", *shared]
            assert arguments == expected

        else:
            expected = ["-m", *shared]
            assert "python" in arguments[0]
            assert arguments[1:] == expected

    @mark_specifiers_less_than_ape
    def test_prepare_install_version_smaller_than_ape(self, specifier, ape_caplog):
        metadata = PluginMetadata(name=list(AVAILABLE_PLUGINS)[0], version=specifier)
        expected = (
            r"Unable to install plugin\.\n"
            r"Reason: Doing so will downgrade Ape's version\.\n"
            r"To resolve: Downgrade Ape first\."
        )
        with pytest.raises(PluginVersionError, match=expected):
            metadata._prepare_install(skip_confirmation=True)

    def test_check_installed_python_39(self, mocker, get_dists_patch):
        mock_dist = mocker.MagicMock()
        get_dists_patch.return_value = [mock_dist]
        mock_dist.name = ""
        mock_dist.metadata = {"Name": "ape-py39plugin"}

        metadata = PluginMetadata(name="py39plugin")

        # If the next line doesn't fail, the test passed.
        # This triggers looping through the dist w/o a name attr.
        assert metadata.check_installed()

    @pytest.mark.parametrize(
        "plugin,expected", [(list(INSTALLED_PLUGINS)[0], True), (list(AVAILABLE_PLUGINS)[0], False)]
    )
    def test_check_installed_python_310_or_greater(self, plugin, expected):
        installed = PluginMetadata(name=plugin)
        assert installed.check_installed() is expected

    @parametrize_pip_cmd
    def test_get_uninstall_args(self, pip_command):
        metadata = PluginMetadata(name="dontmatter", pip_command=pip_command)
        arguments = metadata._get_uninstall_args(python_location=None)
        pip_cmd_len = len(metadata.pip_command)

        for idx, pip_pt in enumerate(pip_command):
            assert arguments[idx] == pip_pt

        expected = ["uninstall"]
        if pip_command[0] == "python":
            expected.append("-y")

        expected.extend(("ape-dontmatter", "--quiet"))
        assert arguments[pip_cmd_len:] == expected


class TestApePluginsRepr:
    def test_str(self, plugin_metadata):
        representation = ApePluginsRepr(plugin_metadata)
        actual = str(representation)
        expected = f"""
Installed Plugins
  installed     {ape_version.base}

Third-party Plugins
  thirdparty    {ape_version.base}
        """
        assert actual == expected.strip()

    def test_str_all_types(self, plugin_metadata):
        representation = ApePluginsRepr(plugin_metadata, include=list(PluginType))
        actual = str(representation)
        expected = f"""
Core Plugins
  run

Installed Plugins
  installed     {ape_version.base}

Third-party Plugins
  thirdparty    {ape_version.base}

Available Plugins
  available
        """
        assert actual == expected.strip()

    def test_str_no_plugins(self):
        plugins = PluginMetadataList.from_package_names([])
        representation = ApePluginsRepr(plugins)
        assert str(representation) == ""


class TestPluginGroup:
    def test_name(self):
        group = PluginGroup(plugin_type=PluginType.INSTALLED)
        assert group.name == "Installed"

    def test_name_when_plugin_type_is_str(self):
        group = PluginGroup(plugin_type=PluginType.INSTALLED)
        group.plugin_type = PluginType.INSTALLED.value  # type: ignore[assignment]
        assert group.name == "Installed"

    def test_repr(self):
        group = PluginGroup(plugin_type=PluginType.INSTALLED)
        assert repr(group) == "<Installed Plugins Group>"

    def test_repr_when_plugin_type_is_str(self):
        group = PluginGroup(plugin_type=PluginType.INSTALLED)
        group.plugin_type = PluginType.INSTALLED.value  # type: ignore[assignment]
        assert repr(group) == "<Installed Plugins Group>"

    def test_repr_when_exception(self, mocker):
        """
        Exceptions CANNOT happen in a repr!
        """
        patch = mocker.patch("ape.plugins._utils.PluginGroup.name", new_callable=mock.PropertyMock)
        patch.side_effect = ValueError("repr fail test")
        group = PluginGroup(plugin_type=PluginType.INSTALLED)

        assert repr(group) == "<PluginGroup>"


def test_handle_upgrade_result_when_upgrading_to_same_version(ape_caplog, logger):
    plugin = PluginMetadata(name=THIRD_PARTY[0], version=f"0.{ape_version.minor}.0")
    handler = ModifyPluginResultHandler(plugin)
    with ape_caplog.at_level(LogLevel.INFO):
        handler.handle_upgrade_result(0, f"0.{ape_version.minor}.0")
        assert (
            f"'{THIRD_PARTY[0]}' already has version '0.{ape_version.minor}.0'" in ape_caplog.head
        )


class TestApeVersion:
    def test_version_range(self):
        actual = ape_version.version_range
        expected = f">=0.{ape_version[2]},<0.{int(ape_version[2]) + 1}"
        assert actual == expected

    def test_next_version_range(self):
        actual = ape_version.next_version_range
        expected = f">=0.{int(ape_version[2]) + 1},<0.{int(ape_version[2]) + 2}"
        assert actual == expected

    def test_previous_version_range(self):
        actual = ape_version.previous_version_range
        expected = f">=0.{int(ape_version[2]) - 2},<0.{int(ape_version[2]) - 1}"
        assert actual == expected

    @mark_specifiers_less_than_ape
    def test_would_be_downgraded(self, specifier):
        assert ape_version.would_get_downgraded(specifier)


def test_filter_plugins_from_dists_py39(mocker):
    def make_dist(name: str):
        mock_dist = mocker.MagicMock()
        mock_dist.name = ""
        mock_dist.metadata = {"Name": f"ape-{name}"}
        return mock_dist

    plugins = [make_dist("solidity"), make_dist("vyper"), make_dist("optimism")]
    actual = set(_filter_plugins_from_dists(plugins))
    expected = {"ape-solidity", "ape-vyper", "ape-optimism"}
    assert actual == expected


def test_filter_plugins_from_dists_py310_and_greater(mocker):
    def make_dist(name: str):
        mock_dist = mocker.MagicMock()
        mock_dist.name = f"ape-{name}"
        return mock_dist

    plugins = [make_dist("solidity"), make_dist("vyper"), make_dist("optimism")]
    actual = set(_filter_plugins_from_dists(plugins))
    expected = {"ape-solidity", "ape-vyper", "ape-optimism"}
    assert actual == expected


@pytest.mark.parametrize("abstract_methods", [("method1", "method2"), {"method1": 0, "method2": 0}])
def test_get_unimplemented_methods_warning_list_containing_plugin(abstract_methods):
    plugin_registration = ("foo", "bar", TransactionAPI)
    actual = _get_unimplemented_methods_warning(plugin_registration, "p1")
    expected = (
        "'TransactionAPI' from 'p1' is not fully implemented. "
        "Remaining abstract methods: 'serialize_transaction, txn_hash'."
    )
    assert actual == expected


def test_core_plugins():
    # In case any of these happen to be installed, and this feature
    # is broken, it will fail. If none are installed, the test will always pass.
    non_core_plugins = ("ape_arbitrum", "ape_vyper", "ape_solidity", "ape_ens")
    assert not any(p in CORE_PLUGINS_LIST for p in non_core_plugins)
    assert "ape_ethereum" in CORE_PLUGINS_LIST


class TestBuildPipArgs:
    @parametrize_pip_cmd
    def test_prepare_package_manager_args_with_python_location(self, pip_command, mock_python_path):
        metadata = PluginMetadata(name="test", pip_command=pip_command)

        args = metadata.prepare_package_manager_args("install", python_location=mock_python_path)

        expected = (
            [*pip_command, "install", "--python", mock_python_path]
            if pip_command[0] == "uv"
            else [*pip_command, "--python", mock_python_path, "install"]
        )
        assert args == expected

    @parametrize_pip_cmd
    def test_prepare_package_manager_args_with_extra_args(self, pip_command):
        metadata = PluginMetadata(name="test", pip_command=pip_command)

        extra = ["--quiet", "--no-deps"]
        args = metadata.prepare_package_manager_args("install", extra_args=extra)

        expected = pip_command + ["install"] + extra
        assert args == expected

    @parametrize_pip_cmd
    def test_prepare_package_manager_args_with_both_options(self, pip_command, mock_python_path):
        metadata = PluginMetadata(name="test", pip_command=pip_command)

        extra = ["--quiet", "--no-deps"]
        args = metadata.prepare_package_manager_args(
            "install", python_location=mock_python_path, extra_args=extra
        )

        expected = (
            [*pip_command, "install", "--python", mock_python_path] + extra
            if pip_command[0] == "uv"
            else [*pip_command, "--python", mock_python_path, "install"] + extra
        )
        assert args == expected

    @parametrize_pip_cmd
    def test_prepare_package_manager_args_with_uninstall_verb(self, pip_command, mock_python_path):
        metadata = PluginMetadata(name="test", pip_command=pip_command)

        args = metadata.prepare_package_manager_args("uninstall", python_location=mock_python_path)

        expected = (
            [*pip_command, "uninstall", "--python", mock_python_path]
            if pip_command[0] == "uv"
            else [*pip_command, "--python", mock_python_path, "uninstall"]
        )
        assert args == expected



import json
from importlib import import_module
from typing import TYPE_CHECKING

import click
from cchecksum import to_checksum_address
from eth_utils import to_hex

from ape.cli import ConnectedProviderCommand
from ape.cli.arguments import existing_alias_argument, non_existing_alias_argument
from ape.cli.options import account_option, ape_cli_context, network_option
from ape.logging import HIDDEN_MESSAGE

if TYPE_CHECKING:
    from ape.api.accounts import AccountAPI
    from ape_accounts.accounts import AccountContainer, KeyfileAccount


ETHEREUM_DEFAULT_PATH = "m/44'/60'/0'/0/0"


def _get_container() -> "AccountContainer":
    from ape.utils.basemodel import ManagerAccessMixin as access

    # NOTE: Must used the instantiated version of `AccountsContainer` in `accounts`
    return access.account_manager.containers["accounts"]


@click.group(short_help="Manage local accounts")
def cli():
    """
    Command-line helper for managing local accounts. You can unlock local accounts from
    scripts or the console using the accounts.load() method.
    """


# Different name because `list` is a keyword
@cli.command(name="list", short_help="List available local accounts")
@ape_cli_context()
@click.option("--all", "show_all_plugins", help="Output accounts from all plugins", is_flag=True)
def _list(cli_ctx, show_all_plugins):
    if "accounts" not in cli_ctx.account_manager.containers:
        cli_ctx.abort("Accounts plugin unexpectedly failed to load.")

    containers = (
        cli_ctx.account_manager.containers if show_all_plugins else {"accounts": _get_container()}
    )
    account_map = {n: [a for a in c.accounts] for n, c in containers.items()}
    account_pairs = [
        pair for pair in {n: ls for n, ls in account_map.items() if len(ls) > 0}.items()
    ]

    if sum(len(c) for c in account_pairs) == 0:
        cli_ctx.logger.warning("No accounts found.")
        return

    num_containers = len(account_pairs)
    for index in range(num_containers):
        plugin_name, container = account_pairs[index]
        num_accounts = len(container)
        header = f"Found {num_accounts} account"
        if num_accounts > 1:
            header = f"{header}s"  # 'account' -> 'accounts'

        if show_all_plugins:
            header = f"{header} in the '{plugin_name}' plugin"

        click.echo(f"{header}:")

        for account in container:
            alias_display = f" (alias: '{account.alias}')" if account.alias else ""
            click.echo(f"  {account.address}{alias_display}")

        if index < num_containers - 1:
            click.echo()


@cli.command(short_help="Create an account with a random mnemonic seed phrase")
@ape_cli_context()
@click.option(
    "--hide-mnemonic",
    help="Hide the newly generated mnemonic from the terminal",
    is_flag=True,
)
@click.option(
    "--word-count",
    help="Number of words to use to generate seed phrase",
    default=12,
    show_default=True,
)
@click.option(
    "--hd-path",
    "custom_hd_path",
    help="Specify an HD path for deriving seed phrase",
    default=ETHEREUM_DEFAULT_PATH,
    show_default=True,
)
@non_existing_alias_argument()
def generate(cli_ctx, alias, hide_mnemonic, word_count, custom_hd_path):
    click.prompt(
        "Enhance the security of your account by adding additional random input",
        hide_input=True,
    )

    show_mnemonic = not hide_mnemonic and click.confirm("Show mnemonic?", default=True)

    passphrase = click.prompt(
        "Create Passphrase to encrypt account",
        hide_input=True,
        confirmation_prompt=True,
    )

    account_module = import_module("ape_accounts.accounts")
    account, mnemonic = account_module.generate_account(
        alias, passphrase, custom_hd_path, word_count
    )

    if show_mnemonic:
        cli_ctx.logger.info(f"Newly generated mnemonic is: {click.style(mnemonic, bold=True)}")

    cli_ctx.logger.success(
        f"A new account '{account.address}' with "
        + f"HDPath {custom_hd_path} has been added with the id '{alias}'"
    )


# Different name because `import` is a keyword
@cli.command(name="import", short_help="Import an account by private key or seed phrase")
@ape_cli_context()
@click.option(
    "--use-mnemonic", "import_from_mnemonic", help="Import a key from a mnemonic", is_flag=True
)
@click.option(
    "--hd-path",
    "custom_hd_path",
    help="Account HD path to use when importing by mnemonic",
    default=ETHEREUM_DEFAULT_PATH,
    show_default=True,
)
@non_existing_alias_argument()
def _import(cli_ctx, alias, import_from_mnemonic, custom_hd_path):
    account: KeyfileAccount | None = None

    def ask_for_passphrase():
        return click.prompt(
            "Create Passphrase to encrypt account",
            hide_input=True,
            confirmation_prompt=True,
        )

    if import_from_mnemonic:
        from eth_account import Account as EthAccount

        mnemonic = click.prompt("Enter mnemonic seed phrase", hide_input=True)
        EthAccount.enable_unaudited_hdwallet_features()
        try:
            passphrase = ask_for_passphrase()
            account = _account_from_mnemonic(alias, passphrase, mnemonic, hd_path=custom_hd_path)
        except Exception as error:
            error_msg = f"{error}".replace(mnemonic, HIDDEN_MESSAGE)
            cli_ctx.abort(f"Seed phrase can't be imported: {error_msg}")

    else:
        key = click.prompt("Enter Private Key", hide_input=True)
        try:
            passphrase = ask_for_passphrase()
            account = _account_from_key(alias, passphrase, key)
        except Exception as error:
            cli_ctx.abort(f"Key can't be imported: {error}")

    if account:
        cli_ctx.logger.success(
            f"A new account '{account.address}' has been added with the id '{alias}'"
        )


def _load_account_type(account: "AccountAPI") -> bool:
    module = import_module("ape_accounts.accounts")
    return isinstance(account, module.KeyfileAccount)


def _account_from_mnemonic(
    alias: str, passphrase: str, mnemonic: str, hd_path: str = ETHEREUM_DEFAULT_PATH
) -> "KeyfileAccount":
    account_module = import_module("ape_accounts.accounts")
    return account_module.import_account_from_mnemonic(alias, passphrase, mnemonic, hd_path=hd_path)


def _account_from_key(alias: str, passphrase: str, key: str) -> "KeyfileAccount":
    account_module = import_module("ape_accounts.accounts")
    return account_module.import_account_from_private_key(alias, passphrase, key)


@cli.command(short_help="Export an account private key")
@ape_cli_context()
@existing_alias_argument(account_type=_load_account_type)
def export(cli_ctx, alias):
    from eth_account import Account as EthAccount

    path = _get_container().data_folder.joinpath(f"{alias}.json")
    account = json.loads(path.read_text(encoding="utf-8"))
    password = click.prompt("Enter password to decrypt account", hide_input=True)
    private_key = EthAccount.decrypt(account, password)
    address = to_checksum_address(account["address"])
    cli_ctx.logger.success(
        f"Account {address} private key: {click.style(to_hex(private_key), bold=True)}"
    )


@cli.command(short_help="Change the password of an existing account")
@ape_cli_context()
@existing_alias_argument(account_type=_load_account_type)
def change_password(cli_ctx, alias):
    account = cli_ctx.account_manager.load(alias)
    account.change_password()
    cli_ctx.logger.success(f"Password has been changed for account '{alias}'")


@cli.command(short_help="Delete an existing account")
@ape_cli_context()
@existing_alias_argument(account_type=_load_account_type)
def delete(cli_ctx, alias):
    account = cli_ctx.account_manager.load(alias)
    account.delete()
    cli_ctx.logger.success(f"Account '{alias}' has been deleted")


@cli.group()
def auth():
    """Manage `EIP-7702` Authorizations for account"""


@auth.command(name="show", cls=ConnectedProviderCommand)
@network_option()
@account_option()
def show_delegate(account):
    """Show if an existing delegate is authorized for account"""
    if contract := account.delegate:
        click.echo(f"{account.address} is delegated to {contract.address}")

    else:
        click.secho(f"{account.address} has no delegate", fg="red")


@auth.command(name="set", cls=ConnectedProviderCommand)
@network_option()
@account_option()
# NOTE: In case users need to override the defaults
@click.option("--receiver", default=None)
@click.option("--data", default=None)
@click.option("--gas-limit", type=int, default=None)
@click.argument("contract")
def authorize_delegate(account, receiver, data, gas_limit, contract):
    """Authorize and set delegate for account"""
    account.set_delegate(contract, receiver=receiver, data=data, gas_limit=gas_limit)
    click.echo(f"{account.address} is now delegated to {contract}")


@auth.command(name="rm", cls=ConnectedProviderCommand)
@network_option()
@account_option()
def remove_delegate(account):
    """Remove delegate for account"""
    account.remove_delegate()



from tests.conftest import geth_process_test


@geth_process_test
def test_contract_event(geth_contract, geth_account):
    geth_contract.setNumber(101010, sender=geth_account)
    actual = geth_contract.NumberChange[-1]
    assert actual.event_name == "NumberChange"
    assert actual.contract_address == geth_contract.address
    assert actual.event_arguments["newNum"] == 101010



from collections.abc import Iterator
from functools import cached_property
from typing import cast

from ape.api.accounts import TestAccountAPI, TestAccountContainerAPI
from ape.exceptions import ProviderNotConnectedError
from ape.types import AddressType
from ape.utils.misc import log_instead_of_fail
from ape.utils.testing import generate_dev_accounts
from ape_accounts.accounts import ApeSigner


class TestAccountContainer(TestAccountContainerAPI):
    generated_accounts: list["TestAccount"] = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __len__(self) -> int:
        return self.number_of_accounts + len(self.generated_accounts)

    @property
    def mnemonic(self) -> str:
        # Overridden so we can overload the setter.
        return self.config_manager.test.mnemonic

    @mnemonic.setter
    def mnemonic(self, mnemonic: str) -> None:
        # Overridden so we can also clear out generated accounts cache.
        self.config_manager.test.mnemonic = mnemonic
        self.generated_accounts = []

    @mnemonic.setter
    def mnemonic(self, mnemonic: str) -> None:
        self.config_manager.test.mnemonic = mnemonic
        self.generated_accounts = []

    @property
    def config(self):
        return self.config_manager.get_config("test")

    @property
    def aliases(self) -> Iterator[str]:
        for index in range(self.number_of_accounts):
            yield f"TEST::{index}"

    @property
    def accounts(self) -> Iterator["TestAccount"]:
        for index in range(self.number_of_accounts):
            yield cast(TestAccount, self.get_test_account(index))

    def get_test_account(self, index: int) -> TestAccountAPI:
        if index >= self.number_of_accounts:
            new_index = index - self.number_of_accounts
            return self.generated_accounts[new_index]

        try:
            return self.provider.get_test_account(index)
        except (NotImplementedError, ProviderNotConnectedError):
            return self.generate_account(index=index)

    def generate_account(self, index: int | None = None) -> "TestAccountAPI":
        new_index = (
            self.number_of_accounts + len(self.generated_accounts) if index is None else index
        )
        generated_account = generate_dev_accounts(
            self.mnemonic, 1, hd_path=self.hd_path, start_index=new_index
        )[0]
        account = self.init_test_account(
            new_index, generated_account.address, generated_account.private_key
        )

        # Only cache if being created outside the expected number of accounts.
        # Else, ends up cached twice and caused logic problems elsewhere.
        if new_index >= self.number_of_accounts:
            self.generated_accounts.append(account)

        return account

    @classmethod
    def init_test_account(cls, index: int, address: AddressType, private_key: str) -> "TestAccount":
        return TestAccount(
            index=index,
            address_str=address,
            private_key=private_key,
        )

    def reset(self):
        self.generated_accounts = []


class TestAccount(ApeSigner, TestAccountAPI):
    index: int
    address_str: str

    __test__ = False

    @property
    def alias(self) -> str:
        return f"TEST::{self.index}"

    @cached_property
    def address(self) -> AddressType:
        # Overridden.
        return self.network_manager.ethereum.decode_address(self.address_str)

    @log_instead_of_fail(default="<TestAccount>")
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}_{self.index} {self.address}>"



import shlex
from functools import cached_property
from pathlib import Path

import click
from click.testing import CliRunner
from IPython import get_ipython
from IPython.core.magic import Magics, line_magic, magics_class
from rich import print as rich_print

import ape
from ape._cli import cli
from ape.exceptions import Abort, ApeException, handle_ape_exception
from ape.logging import logger
from ape.managers.project import LocalProject
from ape.utils.basemodel import ManagerAccessMixin
from ape.utils.os import clean_path


@magics_class
class ApeConsoleMagics(Magics):
    @cached_property
    def ipython(self):
        if ipython := get_ipython():
            return ipython

        raise ValueError("Must be called from an IPython session.")

    @line_magic
    def ape(self, line: str = ""):
        """
        Run Ape CLI commands within an ``ape console`` session.

        Usage example::

            %ape accounts list

        """
        runner = CliRunner()
        if "console" in [x.strip("\"' \t\n") for x in shlex.split(line)]:
            # Prevent running console within console because usually bad
            # stuff happens when you try to do this.
            raise ValueError("Unable to run `console` within a console.")

        result = runner.invoke(cli, line)
        if result.output:
            click.echo(result.output)

        return result

    @line_magic
    def bal(self, line: str = ""):
        """
        Show an account balance in human-readable form.

        Usage example::

            account = accounts.load("me")
            %bal account
        """

        if not line:
            raise ValueError("Missing argument.")

        provider = ape.networks.provider
        ecosystem = provider.network.ecosystem
        result = eval(line, self.ipython.user_global_ns, self.ipython.user_ns)

        if isinstance(result, str) and result.startswith("0x"):
            address = result

        else:
            # Handles accounts, ENS, integers, BaseAddress, and aliases.
            address = ManagerAccessMixin.account_manager.resolve_address(result) or f"{result}"

        decimals = ecosystem.fee_token_decimals
        symbol = ecosystem.fee_token_symbol
        balance = provider.get_balance(address)
        return f"{round(balance / 10**decimals, 8)} {symbol}"


def custom_exception_handler(self, etype, value, tb, tb_offset=None):
    project = self.user_ns["project"]
    if isinstance(project, LocalProject):
        path = project.path
    else:
        # This happens if assigned the variable `project` in your session
        # to something other than ``ape.project``.
        path = ManagerAccessMixin.local_project.path

    if not handle_ape_exception(value, [path]):
        logger.error(Abort.from_ape_exception(value).format_message())


def load_ipython_extension(ipython):
    ipython.register_magics(ApeConsoleMagics)
    ipython.set_custom_exc((ApeException,), custom_exception_handler)

    # This prevents displaying a user's home directory
    # ever when using `ape console`.
    ipython.display_formatter.formatters["text/plain"].for_type(
        Path, lambda x, *args, **kwargs: rich_print(clean_path(x))
    )



import pytest
from eth_pydantic_types import HexBytes
from eth_utils import to_hex

from ape.managers.converters import HexIterableConverter


@pytest.mark.parametrize("val", ("0xA100", "0x0A100", "0x00a100", "00a100"))
def test_hex_str(convert, val):
    assert int(to_hex(convert(val, bytes)), 16) == int(to_hex(0xA100), 16)


@pytest.mark.parametrize("val", ("0xA100", "0x0A100", "0x00a100"))
def test_hex_int(convert, val):
    assert convert(val, int) == 0xA100


@pytest.mark.parametrize(
    "calldata,expected",
    (
        (
            ["0x123456", "0xabcd"],
            "0x123456abcd",
        ),
        (
            [HexBytes("0x123456"), "0xabcd"],
            "0x123456abcd",
        ),
        (
            ("0x123456", "0xabcd"),
            "0x123456abcd",
        ),
    ),
)
def test_hex_concat(calldata, expected, convert):
    assert HexIterableConverter().is_convertible(calldata)
    assert convert(calldata, bytes) == HexBytes(expected)



import pytest

from ape.exceptions import ProviderNotConnectedError
from ape.logging import logger
from ape.managers.project import DependencyManager
from ape.utils.basemodel import DiskCacheableModel, ManagerAccessMixin, only_raise_attribute_error
from ape.utils.os import create_tempdir


class CustomClass(ManagerAccessMixin):
    pass


@pytest.mark.parametrize("accessor", (CustomClass, CustomClass()))
def test_provider(accessor, eth_tester_provider):
    assert accessor.provider == eth_tester_provider


@pytest.mark.parametrize("accessor", (CustomClass, CustomClass()))
def test_provider_not_active(networks, accessor):
    initial = networks.active_provider
    networks.active_provider = None
    try:
        with pytest.raises(ProviderNotConnectedError):
            _ = accessor.provider
    finally:
        networks.active_provider = initial


def test_only_raise_attribute_error(mocker, ape_caplog):
    spy = mocker.spy(logger, "log_debug_stack_trace")

    @only_raise_attribute_error
    def fn():
        raise ValueError("foo bar error")

    with pytest.raises(AttributeError, match="foo bar error"):
        fn()

    assert spy.call_count


def test_only_raise_attribute_error_when_already_raises(mocker, ape_caplog):
    spy = mocker.spy(logger, "log_debug_stack_trace")

    @only_raise_attribute_error
    def fn():
        raise AttributeError("foo bar error")

    with pytest.raises(AttributeError, match="foo bar error"):
        fn()

    # Does not log because is already an attr err
    assert not spy.call_count


def test_dependency_manager():
    actual = ManagerAccessMixin.dependency_manager
    assert isinstance(actual, DependencyManager)


class TestDiskCacheableModel:
    @pytest.fixture(scope="class")
    def ExampleModel(self):
        class _ExampleModel(DiskCacheableModel):
            aa: int
            bb: str
            cc: dict[str, dict[str, int]]

        return _ExampleModel

    def test_model_validate_file(self, ExampleModel):
        with create_tempdir() as path:
            file = path / "example.json"
            json_str = '{"aa":123,"bb":"Hello Pydantic!","cc":{"1":{"2":3}}}'
            file.write_text(json_str)
            instance = ExampleModel.model_validate_file(file)
            file.unlink()

        assert instance.aa == 123
        assert instance.bb == "Hello Pydantic!"
        assert instance.cc == {"1": {"2": 3}}
        # Show the path was already set.
        assert instance._path == file

    def test_model_dump_file(self, ExampleModel):
        instance = ExampleModel(aa=123, bb="Hello Pydantic!", cc={"1": {"2": 3}})
        expected = '{"aa":123,"bb":"Hello Pydantic!","cc":{"1":{"2":3}}}'
        with create_tempdir() as path:
            file = path / "example.json"
            instance.model_dump_file(file)
            actual = file.read_text()

        assert actual == expected



import json
from collections.abc import Iterable, Iterator
from json import JSONDecodeError
from pathlib import Path
from typing import TYPE_CHECKING

from eth_pydantic_types import HexBytes
from eth_utils import is_0x_prefixed
from ethpm_types import ContractType

from ape.api.compiler import CompilerAPI
from ape.exceptions import CompilerError, ContractLogicError
from ape.logging import logger

if TYPE_CHECKING:
    from ape.managers.project import ProjectManager


class InterfaceCompiler(CompilerAPI):
    """
    A compiler plugin for interface JSONs (ABIs). Also, this
    compiler can "compile" already-compiled ``ContractType``
    JSON files.
    """

    @property
    def name(self) -> str:
        return "ethpm"

    def get_versions(self, all_paths: Iterable[Path]) -> set[str]:
        # NOTE: This bypasses the serialization of this compiler into the package manifest's
        #       ``compilers`` field. You should not do this with a real compiler plugin.
        return set()

    def compile(
        self,
        contract_filepaths: Iterable[Path],
        project: "ProjectManager | None",
        settings: dict | None = None,
    ) -> Iterator[ContractType]:
        project = project or self.local_project
        source_ids = {
            p: f"{p.relative_to(project.path)}" if p.is_absolute() else str(p)
            for p in contract_filepaths
        }
        logger.info(f"Compiling {', '.join(source_ids.values())}.")
        for path in contract_filepaths:
            if not path.is_absolute() and (project.path / path).is_file():
                # Was given a relative path to the project.
                src_path = project.path / path

            elif path.is_file():
                # Absolute existing path.
                src_path = path

            else:
                raise CompilerError(f"'{path}' is not a file.")

            code = src_path.read_text()
            source_id = source_ids[path]
            contract_type = self.compile_code(code, project=project, sourceId=source_id)

            # NOTE: Try getting name/ ID from code-JSON first.
            #   That's why this is not part of `**kwargs` in `compile_code()`.
            contract_type.name = contract_type.name or src_path.stem
            yield contract_type

    def compile_code(
        self,
        code: str,
        project: "ProjectManager | None" = None,
        **kwargs,
    ) -> ContractType:
        code = code or "[]"
        try:
            data = json.loads(code)
        except JSONDecodeError as err:
            raise CompilerError(str(err)) from err

        if isinstance(data, list):
            # ABI JSON list
            contract_type_data = {"abi": data, **kwargs}

        elif isinstance(data, dict) and (
            "contractName" in data or "abi" in data or "sourceId" in data
        ):
            # Raw contract type JSON or raw compiler output.
            contract_type_data = {**data, **kwargs}
            if (
                "deploymentBytecode" not in contract_type_data
                or "runtimeBytecode" not in contract_type_data
            ):
                if "bin" in contract_type_data:
                    # Handle raw Solidity output.
                    deployment_bytecode = data["bin"]
                    runtime_bytecode = data["bin"]

                elif "bytecode" in contract_type_data or "bytecode_runtime" in contract_type_data:
                    # Handle raw Vyper output.
                    deployment_bytecode = contract_type_data.pop("bytecode", None)
                    runtime_bytecode = contract_type_data.pop("bytecode_runtime", None)

                else:
                    deployment_bytecode = None
                    runtime_bytecode = None

                if deployment_bytecode:
                    contract_type_data["deploymentBytecode"] = {"bytecode": deployment_bytecode}
                if runtime_bytecode:
                    contract_type_data["runtimeBytecode"] = {"bytecode": runtime_bytecode}

        else:
            raise CompilerError(f"Unable to parse {ContractType.__name__}.")

        return ContractType(**contract_type_data)

    def enrich_error(self, err: ContractLogicError) -> ContractLogicError:
        if not (address := err.address) or not is_0x_prefixed(err.revert_message):
            return err

        # Check for ErrorABI.
        bytes_message = HexBytes(err.revert_message)
        selector = bytes_message[:4]
        input_data = bytes_message[4:]

        try:
            contract = self.chain_manager.contracts.instance_at(address)
        except Exception:
            return err

        if (
            not contract
            or not self.network_manager.active_provider
            or selector not in contract.contract_type.errors
        ):
            return err

        ecosystem = self.provider.network.ecosystem
        abi = contract.contract_type.errors[selector]
        inputs = ecosystem.decode_calldata(abi, input_data)
        error_class = contract.get_error_by_signature(abi.signature)
        return error_class(
            abi,
            inputs,
            txn=err.txn,
            trace=lambda: err.trace,
            contract_address=address,
            source_traceback=lambda: err.source_traceback,
        )



import sys

import pytest

from tests.conftest import ApeSubprocessRunner

from .utils import skip_projects_except

BAD_COMMAND = "not-a-name"


@pytest.fixture
def scripts_runner(config):
    class ScriptsSubprocessRunner(ApeSubprocessRunner):
        def __init__(self):
            super().__init__("run", data_folder=config.DATA_FOLDER)

    return ScriptsSubprocessRunner()


@skip_projects_except("script")
def test_run_unknown_script(scripts_runner, integ_project):
    scripts_runner.project = integ_project
    result = scripts_runner.invoke(BAD_COMMAND)
    assert result.exit_code == 2
    assert f"No such command '{BAD_COMMAND}'." in result._completed_process.stderr


@skip_projects_except("script")
def test_run(scripts_runner, integ_project):
    scripts_runner.project = integ_project
    result = scripts_runner.invoke()
    assert result.exit_code == 0, result.output
    # By default, no commands are run
    assert "Super secret script output" not in result.output

    not_part_of_test = ("output_contract_view_methods",)
    scripts = [
        s
        for s in integ_project.scripts_folder.glob("*.py")
        if not s.name.startswith("error") and s.stem not in not_part_of_test
    ]
    for script_file in scripts:
        result = scripts_runner.invoke(script_file.stem)
        assert result.exit_code == 0, (
            f"Unexpected exit code for '{script_file.name}'\n{result.output}"
        )

        if script_file.stem.startswith("_"):
            assert "Super secret script output" not in result.output

        else:
            assert "Super secret script output" in result.output


@skip_projects_except("script")
def test_run_with_verbosity(scripts_runner, integ_project):
    scripts_runner.project = integ_project
    result = scripts_runner.invoke("click", "--verbosity", "DEBUG")
    assert result.exit_code == 0, result.output or result._completed_process.stderr


@skip_projects_except("script")
def test_run_subdirectories(scripts_runner, integ_project):
    scripts_runner.project = integ_project
    result = scripts_runner.invoke()
    assert result.exit_code == 0, result.output
    # By default, no commands are run
    assert "Super secret script output" not in result.output
    subdirectory_scripts = [
        s
        for s in (integ_project.scripts_folder / "subdirectory").rglob("*.py")
        if not s.name.startswith("error")
    ]
    for each in subdirectory_scripts:
        result = scripts_runner.invoke("subdirectory", each.stem)
        assert result.exit_code == 0
        assert "Super secret script output" in result.output


@skip_projects_except("only-script-subdirs")
def test_run_only_subdirs(scripts_runner, integ_project):
    scripts_runner.project = integ_project
    result = scripts_runner.invoke()
    assert result.exit_code == 0, result.output
    # By default, no commands are run
    assert "Super secret script output" not in result.output
    subdirectory_scripts = [
        s
        for s in (integ_project.scripts_folder / "subdirectory").rglob("*.py")
        if not s.name.startswith("error")
    ]
    for each in subdirectory_scripts:
        result = scripts_runner.invoke("subdirectory", each.stem)
        assert result.exit_code == 0, f"Unexpected exit code for '{each.name}'"
        assert "Super secret script output" in result.output


@skip_projects_except("script")
def test_run_when_script_errors(scripts_runner, integ_project):
    scripts_runner.project = integ_project
    scripts = [
        s
        for s in integ_project.scripts_folder.glob("*.py")
        if s.name.startswith("error") and not s.name.endswith("forgot_click.py")
    ]
    for script_file in scripts:
        result = scripts_runner.invoke(
            script_file.stem,
        )
        assert result.exit_code != 0, (
            f"Unexpected exit code for '{script_file.name}'.\n{result.output}"
        )
        assert "Expected exception" in result._completed_process.stderr


@skip_projects_except("script")
def test_run_interactive(scripts_runner, integ_project):
    scripts_runner.project = integ_project
    error_names = ("error_main", "error_cli", "error_no_def")
    scripts = [integ_project.scripts_folder / f"{s}.py" for s in error_names]

    # Show that the variable namespace from the script is available in the console.
    user_input = (
        "local_variable\nape.chain.provider.mine()\n"
        "print(f'timestamp={ape.chain.blocks.head.timestamp}')\nexit\n"
    )

    result = scripts_runner.invoke("--interactive", scripts[0].stem, input=user_input)
    assert result.exit_code == 0, result.output

    # From script: local_variable = "test foo bar"
    assert "test foo bar" in result.output
    assert "timestamp=123123123123123" in result.output


@skip_projects_except("script")
def test_run_custom_provider(scripts_runner, integ_project):
    scripts_runner.project = integ_project
    result = scripts_runner.invoke("deploy", "--network", "ethereum:mainnet:http://127.0.0.1:9545")

    # Show that it attempts to connect
    assert result.exit_code == 1, result.output
    assert "No (supported) node found on 'http://127.0.0.1:9545" in result.output


@skip_projects_except("script")
def test_run_custom_network(scripts_runner, integ_project):
    scripts_runner.project = integ_project
    result = scripts_runner.invoke("deploy", "--network", "http://127.0.0.1:9545")

    # Show that it attempts to connect
    assert result.exit_code == 1, result.output
    assert "No (supported) node found on 'http://127.0.0.1:9545" in result.output


@skip_projects_except("script")
def test_try_run_script_missing_cli_decorator(scripts_runner, integ_project):
    """
    Shows that we cannot run a script defining a `cli()` method without
    it being a click command. The script is not recognized, so you get
    a usage error.
    """
    scripts_runner.project = integ_project
    result = scripts_runner.invoke("error_forgot_click")
    assert "Usage: ape run" in result._completed_process.stderr


@skip_projects_except("with-contracts")
def test_uncaught_tx_err(scripts_runner, integ_project):
    scripts_runner.project = integ_project
    result = scripts_runner.invoke("txerr")
    assert '/scripts/txerr.py", line 12, in main' in result.output
    assert "contract.setNumber(5, sender=account)" in result.output
    assert "ERROR" in result.output
    assert "(ContractLogicError) Transaction failed." in result.output


@skip_projects_except("script")
def test_scripts_module_already_installed(integ_project, scripts_runner, mocker):
    """
    Make sure that if there is for some reason a python module names `scripts`
    installed, it does not interfere with Ape's scripting mechanism.
    """
    scripts_runner.project = integ_project
    mock_scripts = mocker.MagicMock()
    mock_path = mocker.MagicMock()
    mock_path._path = "path/to/scripts"
    mock_scripts.__file__ = None
    mock_scripts.__path__ = mock_path
    sys.modules["scripts"] = mock_scripts
    result = scripts_runner.invoke()
    assert result.exit_code == 0, result.output
    del sys.modules["scripts"]


@skip_projects_except("script")
def test_run_recompiles_if_needed(runner, ape_cli, scripts_runner, integ_project):
    """
    Ensure that when a change is made to a contract,
    when we run a script, it re-compiles the script first.
    """
    scripts_runner.project = integ_project

    # Ensure we begin compiled.
    runner.invoke(ape_cli, ("compile", "--force", "--project", f"{integ_project.path}"))

    # Make a change to the contract.
    contract = integ_project.contracts_folder / "VyperContract.json"
    method_name = integ_project.VyperContract.contract_type.view_methods[0].name
    new_method_name = f"f__{method_name}__"
    new_contract_text = contract.read_text().replace(method_name, new_method_name)
    contract.write_text(new_contract_text, encoding="utf8")

    # Run the script. It better recompile first!
    result = scripts_runner.invoke("output_contract_view_methods")
    assert new_method_name in result.output



from enum import Enum


class Scope(int, Enum):
    SESSION = 0
    PACKAGE = 1
    MODULE = 2
    CLASS = 3
    FUNCTION = 4

    def __str__(self) -> str:
        return self.name.lower()

    @property
    def isolation_fixturename(self) -> str:
        return f"_{self}_isolation"



from collections.abc import Sequence
from fnmatch import fnmatch
from statistics import mean, median
from typing import TYPE_CHECKING, Any

from eth_utils import is_0x_prefixed, to_hex
from rich.box import SIMPLE
from rich.table import Table

from ape.utils.misc import is_evm_precompile, is_zero_hex

if TYPE_CHECKING:
    from ape.types.coverage import CoverageReport
    from ape.types.trace import ContractFunctionPath, GasReport

USER_ASSERT_TAG = "USER_ASSERT"
DEFAULT_WRAP_THRESHOLD = 50


def prettify_function(
    method: str,
    calldata: Any,
    contract: str | None = None,
    returndata: Any | None = None,
    stylize: bool = False,
    is_create: bool = False,
    depth: int = 0,
) -> str:
    """
    Prettify the given method call-string to a displayable, prettier string.
    Useful for displaying traces and decoded calls.

    Args:
        method (str): the method call-string to prettify.
        calldata (Any): Arguments to the method.
        contract (str | None): The contract name called.
        returndata (Any): Returned values from the method.
        stylize (bool): ``True`` to use rich styling.
        is_create (bool): Set to ``True`` if creating a contract for better styling.
        depth (int): The depth in the trace (or output) this function gets displayed.

    Returns:
        str
    """
    if "(" in method:
        # Only show short name, not ID name
        # (it is the full signature when multiple methods have the same name).
        method = method.split("(")[0].strip() or method

    if stylize:
        method = f"[{TraceStyles.METHODS}]{method}[/]"
        if contract:
            contract = f"[{TraceStyles.CONTRACTS}]{contract}[/]"

    arguments_str = prettify_inputs(calldata, stylize=stylize)
    if is_create and is_0x_prefixed(arguments_str):
        # Un-enriched CREATE calldata is a massive hex.
        arguments_str = "()"

    signature = f"{method}{arguments_str}"
    if not is_create and returndata not in ((), [], None, {}, ""):
        if return_str := _get_outputs_str(returndata, stylize=stylize, depth=depth):
            signature = f"{signature} -> {return_str}"

    if contract:
        signature = f"{contract}.{signature}"

    return signature


def prettify_inputs(inputs: Any, stylize: bool = False) -> str:
    """
    Prettify the inputs to a function or event (or alike).

    Args:
        inputs (Any): the inputs to prettify.
        stylize (bool): ``True`` to use rich styling.

    Returns:
        str
    """
    color = TraceStyles.INPUTS if stylize else None
    if inputs in ["0x", None, (), [], {}]:
        return "()"

    elif isinstance(inputs, dict):
        return prettify_dict(inputs, color=color)

    elif isinstance(inputs, bytes):
        return to_hex(inputs)

    return f"({inputs})"


def _get_outputs_str(outputs: Any, stylize: bool = False, depth: int = 0) -> str | None:
    if outputs in ["0x", None, (), [], {}]:
        return None

    elif isinstance(outputs, dict):
        color = TraceStyles.OUTPUTS if stylize else None
        return prettify_dict(outputs, color=color)

    elif isinstance(outputs, (list, tuple)):
        return (
            f"[{TraceStyles.OUTPUTS}]{prettify_list(outputs)}[/]"
            if stylize
            else prettify_list(outputs, depth=depth)
        )

    return f"[{TraceStyles.OUTPUTS}]{outputs}[/]" if stylize else str(outputs)


def prettify_list(
    ls: list | tuple,
    depth: int = 0,
    indent: int = 2,
    wrap_threshold: int = DEFAULT_WRAP_THRESHOLD,
) -> str:
    """
    Prettify a list of values for displaying.

    Args:
        ls (list): the list to prettify.
        depth (int): The depth the list appears in a tree structure (for traces).

    Returns:
        str
    """
    if not isinstance(ls, (list, tuple)) or len(str(ls)) < wrap_threshold:
        return str(ls)

    elif ls and isinstance(ls[0], (list, tuple)):
        # List of lists
        sub_lists = [prettify_list(i) for i in ls]

        # Use multi-line if exceeds threshold OR any of the sub-lists use multi-line
        extra_chars_len = (len(sub_lists) - 1) * 2
        use_multiline = len(str(sub_lists)) + extra_chars_len > wrap_threshold or any(
            ["\n" in ls for ls in sub_lists]
        )

        if not use_multiline:
            # Happens for lists like '[[0], [1]]' that are short.
            return f"[{', '.join(sub_lists)}]"

        value = "[\n"
        num_sub_lists = len(sub_lists)
        index = 0
        spacing = indent * " " * 2
        for formatted_list in sub_lists:
            if "\n" in formatted_list:
                # Multi-line sub list. Append 1 more spacing to each line.
                indented_item = f"\n{spacing}".join(formatted_list.splitlines())
                value = f"{value}{spacing}{indented_item}"
            else:
                # Single line sub-list
                value = f"{value}{spacing}{formatted_list}"

            if index < num_sub_lists - 1:
                value = f"{value},"

            value = f"{value}\n"
            index += 1

        value = f"{value}]"
        return value

    return _list_to_multiline_str(ls, depth=depth)


def prettify_dict(
    dictionary: dict,
    color: str | None = None,
    indent: int = 2,
    wrap_threshold: int = DEFAULT_WRAP_THRESHOLD,
) -> str:
    """
    Prettify a dictionary.

    Args:
        dictionary (dict): The dictionary to prettify.
        color (str | None): The color to use for pretty printing.

    Returns:
        str
    """
    length = sum(len(str(v)) for v in [*dictionary.keys(), *dictionary.values()])
    do_wrap = length > wrap_threshold

    index = 0
    end_index = len(dictionary) - 1
    kv_str = "(\n" if do_wrap else "("

    for key, value in dictionary.items():
        if do_wrap:
            kv_str += indent * " "

        if isinstance(value, (list, tuple)):
            value = prettify_list(value, 1 if do_wrap else 0)

        value_str = f"[{color}]{value}[/]" if color is not None else str(value)
        kv_str += f"{key}={value_str}" if key and not key.isnumeric() else value_str
        if index < end_index:
            kv_str += ", "

        if do_wrap:
            kv_str += "\n"

        index += 1

    return f"{kv_str})"


def _list_to_multiline_str(value: list | tuple, depth: int = 0, indent: int = 2) -> str:
    spacing = indent * " "
    ls_spacing = spacing * (depth + 1)
    joined = ",\n".join([f"{ls_spacing}{v}" for v in value])
    new_val = f"[\n{joined}\n{spacing * depth}]"
    return new_val


class TraceStyles:
    """
    Colors to use when displaying a call trace.
    Each item in the class points to the part of
    the trace it colors.
    """

    CONTRACTS = "#ff8c00"
    """Contract type names."""

    METHODS = "bright_green"
    """Method names; not including arguments or return values."""

    INPUTS = "bright_magenta"
    """Method arguments."""

    OUTPUTS = "bright_blue"
    """Method return values."""

    DELEGATE = "#d75f00"
    """The part '(delegate)' that appears before delegate calls."""

    VALUE = "#00afd7"
    """The transaction value, when it's > 0."""

    GAS_COST = "dim"
    """The gas used of the call."""


def parse_gas_table(report: "GasReport") -> list[Table]:
    tables: list[Table] = []

    for contract_id, method_calls in report.items():
        title = f"{contract_id} Gas"
        table = Table(title=title, box=SIMPLE)
        table.add_column("Method")
        table.add_column("Times called", justify="right")
        table.add_column("Min.", justify="right")
        table.add_column("Max.", justify="right")
        table.add_column("Mean", justify="right")
        table.add_column("Median", justify="right")
        has_at_least_1_row = False

        for method_call, gases in sorted(method_calls.items()):
            if not gases:
                continue

            if not method_call or is_zero_hex(method_call) or is_evm_precompile(method_call):
                continue

            elif method_call == "__new__":
                # Looks better in the gas report.
                method_call = "__init__"

            has_at_least_1_row = True
            table.add_row(
                method_call,
                f"{len(gases)}",
                f"{min(gases)}",
                f"{max(gases)}",
                f"{int(round(mean(gases)))}",
                f"{int(round(median(gases)))}",
            )

        if has_at_least_1_row:
            tables.append(table)

    return tables


def parse_coverage_tables(
    coverage: "CoverageReport", verbose: bool = False, statement: bool = True
) -> list[Table]:
    return (
        _parse_verbose_coverage(coverage, statement=statement)
        if verbose
        else [_parse_coverage_table(coverage, statement=statement)]
    )


def _parse_coverage_table(coverage: "CoverageReport", statement: bool = True) -> Table:
    table = Table(title="Contract Coverage", box=SIMPLE)

    # NOTE: Purposely uses same column names as coveragepy
    table.add_column("Name")
    if statement:
        table.add_column("Stmts", justify="right")
        table.add_column("Miss", justify="right")
        table.add_column("Cover", justify="right")

    table.add_column("Funcs", justify="right")

    for project in coverage.projects:
        for src in sorted(project.sources, key=lambda x: x.source_id.lower()):
            row = (
                (
                    src.source_id,
                    f"{src.lines_valid}",
                    f"{src.miss_count}",
                    f"{round(src.line_rate * 100, 2)}%",
                    f"{round(src.function_rate * 100, 2)}%",
                )
                if statement
                else (src.source_id, f"{round(src.function_rate * 100, 2)}%")
            )
            table.add_row(*row)

    return table


def _parse_verbose_coverage(coverage: "CoverageReport", statement: bool = True) -> list[Table]:
    tables = []
    row: tuple[str, ...]
    for project in coverage.projects:
        for src in project.sources:
            for contract in src.contracts:
                title = f"{contract.name} Coverage"
                line_rate = round(contract.line_rate * 100, 2)
                fn_rate = round(contract.function_rate * 100, 2)
                caption = f"line={line_rate}%, func={fn_rate}%"
                table = Table(title=title, box=SIMPLE, caption=caption)
                rows: list[tuple[str, ...]] = []
                table.add_column("Func", justify="right")

                if statement:
                    table.add_column("Stmts", justify="right")
                    table.add_column("Miss", justify="right")

                table.add_column("Cover", justify="right")
                for fn in contract.functions:
                    if fn.name == "__builtin__" and not statement:
                        # Ignore builtins when statement coverage is not being asked for.
                        # It is impossible to really track.
                        continue

                    if fn.name == "__builtin__":
                        # Create a row per unique type.
                        builtins = {x.tag for x in fn.statements if x.tag}
                        for builtin in builtins:
                            name_chars = [
                                c
                                for c in builtin.lower().strip().replace(" ", "_")
                                if c.isalpha() or c == "_"
                            ]
                            name = f"__{''.join(name_chars).replace('dev_', '')}__"
                            miss = (
                                0
                                if any(s.hit_count > 0 for s in fn.statements if s.tag == builtin)
                                else 1
                            )
                            rows.append(
                                tuple((name, name, "1", f"{miss}", "0.0%" if miss else "100.0%"))
                            )

                    else:
                        row = (
                            (
                                fn.name,
                                fn.full_name,
                                f"{fn.lines_valid}",
                                f"{fn.miss_count}",
                                f"{round(fn.line_rate * 100, 2)}%",
                            )
                            if statement
                            else (fn.name, fn.full_name, "✓" if fn.hit_count > 0 else "x")
                        )
                        rows.append(row)

                # Handle cases where normal names are duplicated.
                # Use full names in this case.
                rows_corrected = []
                while rows:
                    row = rows.pop()
                    if row[0] in [r[0] for r in rows]:
                        # Use full-name for all with same name.
                        rows_corrected.append(row[1:])
                        for subrow in rows:
                            if subrow[0] != row[0]:
                                continue

                            rows_corrected.append(subrow[1:])
                            rows.remove(subrow)

                    else:
                        # Use smaller name (no duplicates).
                        rows_corrected.append((row[0], *row[2:]))

                for tbl_row in sorted(rows_corrected):
                    table.add_row(*tbl_row)

                tables.append(table)

    return tables


def _exclude_gas(
    exclusions: Sequence["ContractFunctionPath"], contract_id: str, method_id: str
) -> bool:
    for exclusion in exclusions:
        if exclusion.method_name is None and fnmatch(contract_id, exclusion.contract_name):
            # Skip this whole contract. Search contracts from sub-calls.
            return True

        for excl in exclusions:
            if not excl.method_name:
                # Full contract skips handled above.
                continue

            elif not fnmatch(contract_id, excl.contract_name):
                # Method may match, but contract does not match, so continue.
                continue

            elif method_id and fnmatch(method_id, excl.method_name):
                # Skip this report because of the method name exclusion criteria.
                return True

    return False



from collections.abc import Iterator
from typing import TYPE_CHECKING

from eth_account import Account
from eth_account.messages import SignableMessage
from eth_pydantic_types import HexBytes
from eth_utils import to_bytes, to_hex
from pydantic.dataclasses import dataclass

from ape.utils.misc import as_our_module, log_instead_of_fail

if TYPE_CHECKING:
    from ape.types.address import AddressType

    try:
        # Only on Python 3.11
        from typing import Self  # type: ignore
    except ImportError:
        from typing_extensions import Self  # type: ignore


# Fix 404 in doc link.
as_our_module(
    SignableMessage,
    doc_str=(SignableMessage.__doc__ or "").replace(
        "EIP-191_", "`EIP-191 <https://eips.ethereum.org/EIPS/eip-191>`__"
    ),
)


# Improve repr to force hexstr for body instead of raw bytes.
def signable_message_repr(msg) -> str:
    name = getattr(SignableMessage, "__name__", "SignableMessage")
    default_value = "<unknown!>"  # Shouldn't happen
    version_str = _bytes_to_human_str(msg.version) or default_value
    header_str = _bytes_to_human_str(msg.header) or default_value
    body_str = _bytes_to_human_str(msg.body) or default_value
    return f'{name}(version="{version_str}", header="{header_str}", body="{body_str}")'


SignableMessage.__repr__ = signable_message_repr  # type: ignore[method-assign]


def _bytes_to_human_str(bytes_value: bytes) -> str | None:
    try:
        # Try as text
        return bytes_value.decode("utf8")
    except Exception:
        pass

    try:
        # Try as hex
        return to_hex(bytes_value)
    except Exception:
        pass

    try:
        # Try normal str
        return str(bytes_value)
    except Exception:
        pass

    return None


def _left_pad_bytes(val: bytes, num_bytes: int) -> bytes:
    return b"\x00" * (num_bytes - len(val)) + val if len(val) < num_bytes else val


@dataclass(frozen=True)
class _Signature:
    v: int
    """
    The version byte (``v``) in an Ethereum-style ECDSA signature.
    """

    r: bytes
    """
    The random point (``r``) in an ECDSA signature.
    """

    s: bytes
    """
    The signature proof point (``s``) in an ECDSA signature.
    """

    def __iter__(self) -> Iterator[int | bytes]:
        # NOTE: Allows tuple destructuring
        yield self.v
        yield self.r
        yield self.s

    @classmethod
    def from_rsv(cls, rsv: HexBytes) -> "Self":
        # NOTE: Values may be padded.
        if len(rsv) != 65:
            raise ValueError("Length of RSV bytes must be 65.")

        return cls(r=HexBytes(rsv[:32]), s=HexBytes(rsv[32:64]), v=rsv[64])

    @classmethod
    def from_vrs(cls, vrs: HexBytes) -> "Self":
        # NOTE: Values may be padded.
        if len(vrs) != 65:
            raise ValueError("Length of VRS bytes must be 65.")

        return cls(v=vrs[0], r=HexBytes(vrs[1:33]), s=HexBytes(vrs[33:]))

    @log_instead_of_fail(default="<_Signature>")
    def __repr__(self) -> str:
        class_name = getattr(type(self), "__name__", "_Signature")
        return f"<{class_name} v={self.v} r={to_hex(self.r)} s={to_hex(self.s)}>"

    def encode_vrs(self) -> bytes:
        return to_bytes(self.v) + _left_pad_bytes(self.r, 32) + _left_pad_bytes(self.s, 32)

    def encode_rsv(self) -> bytes:
        return _left_pad_bytes(self.r, 32) + _left_pad_bytes(self.s, 32) + to_bytes(self.v)


class MessageSignature(_Signature):
    """
    A ECDSA signature (vrs) of a message.
    """


def recover_signer(msg: SignableMessage, sig: MessageSignature) -> "AddressType":
    """
    Get the address of the signer.

    Args:
        msg (:class:`~ape.types.signatures.SignableMessage`): A formatted and signable
          message.
        sig (:class:`~ape.types.signatures.MessageSignature`): Signature of the message.

    Returns:
        :class:`~ape.types.address.AddressType`: address of message signer.
    """
    return Account.recover_message(msg, sig)


class TransactionSignature(_Signature):
    """
    A ECDSA signature (vrs) of a transaction.
    """


__all__ = ["MessageSignature", "SignableMessage", "TransactionSignature"]



"""Utilities to extract debug log data from Vyper's print() and the Hardhat Solidity console.log
calls.

There are two different implementations of the console.log functionality. Vyper's print() and
Hardhat's console.sol contract. Both send staticcalls to a specific non-existent contract that we
can read from the call tree.

Vyper encodes these calls as `log(string,bytes)`. Where the first argument is an ABI signature
(e.g. "(uint256,uint8,string)") and the second arg is the printed data. This allows for a simple ABI
and very dynamic output.

Hardhat's console.sol has a strict ABI (unless patched :-/) that mixes and matches all the basic
datatypes to overload all calls to `log()`.

References
----------
- Vyper `print()`: https://docs.vyperlang.org/en/latest/built-in-functions.html#print
- Hardhat's `console.sol`: https://github.com/NomicFoundation/hardhat/blob/e6a42cf330863a6c64624baedd82fc6c5218a20a/packages/hardhat-core/console.sol#L4  # noqa: E501
- Discussion on dynamic ABI encoding (Vyper-style) for log calls: https://github.com/NomicFoundation/hardhat/issues/2666  # noqa: E501
"""

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, cast

from eth_abi import decode
from eth_typing import ChecksumAddress
from eth_utils import add_0x_prefix, decode_hex, to_hex
from ethpm_types import ContractType, MethodABI
from hexbytes import HexBytes

import ape
from ape_ethereum._console_log_abi import CONSOLE_LOG_ABI

if TYPE_CHECKING:
    from typing import TypeGuard

    from evm_trace import CallTreeNode


CONSOLE_ADDRESS = cast(ChecksumAddress, "0x000000000000000000636F6e736F6c652e6c6f67")
VYPER_PRINT_METHOD_ID = HexBytes("0x23cdd8e8")  # log(string,bytes)

console_contract = ContractType(abi=CONSOLE_LOG_ABI, contractName="console")


def is_console_log(call: "CallTreeNode") -> "TypeGuard[CallTreeNode]":
    """Determine if a call is a standard console.log() call"""
    return (
        call.address == HexBytes(CONSOLE_ADDRESS)
        and to_hex(call.calldata[:4]) in console_contract.identifier_lookup
    )


def is_vyper_print(call: "CallTreeNode") -> "TypeGuard[CallTreeNode]":
    """Determine if a call is a standard Vyper print() call"""
    if call.address != HexBytes(CONSOLE_ADDRESS) or call.calldata[:4] != VYPER_PRINT_METHOD_ID:
        return False

    schema, _ = decode(["string", "bytes"], call.calldata[4:])
    types = schema.strip("()").split(",")

    # Now we look at the first arg to try and determine if it's an ABI signature
    # TODO: Tighten this up. This is not entirely accurate, but should mostly get us there.
    return len(types) > 0 and (
        types[0].startswith("uint")
        or types[0].startswith("int")
        or types[0].startswith("bytes")
        or types[0] == "string"
    )


def console_log(method_abi: MethodABI, calldata: str) -> tuple[Any]:
    """Return logged data for console.log() calls"""
    bcalldata = decode_hex(calldata)
    data = ape.networks.ethereum.decode_calldata(method_abi, bcalldata)
    return tuple(data.values())


def vyper_print(calldata: str) -> tuple[Any]:
    """Return logged data for print() calls"""
    schema, payload = decode(["string", "bytes"], HexBytes(calldata))
    data = decode(schema.strip("()").split(","), payload)
    return tuple(data)


def extract_debug_logs(call: "CallTreeNode") -> Iterable[tuple[Any]]:
    """Filter calls to console.log() and print() from a transactions call tree"""
    if is_vyper_print(call) and call.calldata is not None:
        yield vyper_print(add_0x_prefix(to_hex(call.calldata[4:])))

    elif is_console_log(call) and call.calldata is not None:
        method_abi = console_contract.identifier_lookup.get(to_hex(call.calldata[:4]))
        if isinstance(method_abi, MethodABI):
            yield console_log(method_abi, to_hex(call.calldata[4:]))

    elif call.calls is not None:
        for sub_call in call.calls:
            yield from extract_debug_logs(sub_call)



import click
import pytest
from click.testing import CliRunner

from ape.cli import ape_cli_context
from ape.logging import LogLevel, logger, sanitize_url, silenced


@pytest.fixture
def simple_runner():
    return CliRunner()


@click.group()
def group_for_testing():
    pass


def test_info(simple_runner):
    @group_for_testing.command()
    @ape_cli_context()
    def cmd(cli_ctx):
        with cli_ctx.logger.at_level(LogLevel.INFO):
            cli_ctx.logger.info("this is a test")

    result = simple_runner.invoke(group_for_testing, "cmd")
    assert "INFO" in result.output
    assert "this is a test" in result.output


def test_info_level_higher(simple_runner):
    @group_for_testing.command()
    @ape_cli_context()
    def cmd(cli_ctx):
        cli_ctx.logger.info("this is a test")

    logger._did_parse_sys_argv = False
    result = simple_runner.invoke(group_for_testing, ("cmd", "-v", "WARNING"))

    # You don't get INFO log when log level is higher
    assert "INFO" not in result.output
    assert "this is a test" not in result.output


def test_warning(simple_runner):
    @group_for_testing.command()
    @ape_cli_context()
    def cmd(cli_ctx):
        with cli_ctx.logger.at_level(LogLevel.WARNING):
            cli_ctx.logger.warning("this is a test")

    result = simple_runner.invoke(group_for_testing, "cmd")
    assert "WARNING" in result.output
    assert "this is a test" in result.output


def test_warning_level_higher(simple_runner):
    @group_for_testing.command()
    @ape_cli_context()
    def cmd(cli_ctx):
        with cli_ctx.logger.at_level(LogLevel.WARNING.value + 1):
            cli_ctx.logger.warning("this is a test")

    logger._did_parse_sys_argv = False
    result = simple_runner.invoke(group_for_testing, ("cmd", "-v", "ERROR"))
    assert "WARNING" not in result.output
    assert "this is a test" not in result.output


def test_success(simple_runner):
    # Since the log level defaults to INFO,
    # this test also ensures that we get SUCCESS logs
    # without having to specify verbosity
    @group_for_testing.command()
    @ape_cli_context(default_log_level=LogLevel.INFO.value)
    def cmd(cli_ctx):
        with cli_ctx.logger.at_level(LogLevel.SUCCESS):
            cli_ctx.logger.success("this is a test")

    logger._did_parse_sys_argv = False
    result = simple_runner.invoke(group_for_testing, "cmd")
    assert "SUCCESS" in result.output
    assert "this is a test" in result.output


def test_success_level_higher(simple_runner):
    @group_for_testing.command()
    @ape_cli_context()
    def cmd(cli_ctx):
        with cli_ctx.logger.at_level(LogLevel.SUCCESS.value + 1):
            cli_ctx.logger.success("this is a test")

    logger._did_parse_sys_argv = False
    result = simple_runner.invoke(group_for_testing, ("cmd", "-v", "WARNING"))
    assert "SUCCESS" not in result.output
    assert "this is a test" not in result.output


def test_format(simple_runner):
    @group_for_testing.command()
    @ape_cli_context()
    def cmd(cli_ctx):
        cli_ctx.logger.format(fmt="%(message)s")
        try:
            with cli_ctx.logger.at_level(LogLevel.SUCCESS):
                cli_ctx.logger.success("this is a test")
        finally:
            cli_ctx.logger.format()

    logger._did_parse_sys_argv = False
    result = simple_runner.invoke(group_for_testing, ("cmd", "-v", "SUCCESS"))
    assert "SUCCESS" not in result.output


@pytest.mark.parametrize(
    "level", (LogLevel.INFO, LogLevel.INFO.value, LogLevel.INFO.name, "LogLevel.INFO")
)
def test_set_level(level):
    logger.set_level(level)
    assert logger.level == LogLevel.INFO.value


def test_at_level():
    level_to_set = next(lvl for lvl in LogLevel if lvl != logger.level)
    initial_level = logger.level
    with logger.at_level(level_to_set):
        assert logger.level == level_to_set

    assert logger.level == initial_level


@pytest.mark.parametrize(
    "url", ("https://user:password@example.com/v1/API_KEY", "https://example.com/v1/API_KEY")
)
def test_sanitize_url(url):
    actual = sanitize_url(url)
    expected = "https://example.com/[hidden]"
    assert actual == expected


def test_disabled(ape_caplog):
    message = "Can you hear me now?"
    with logger.disabled():
        logger.error(message)
        assert message not in ape_caplog.head

    # Show it is back.
    logger.error(message)
    assert message in ape_caplog.head


def test_silenced(ape_caplog):
    @silenced
    def method_to_silence(x: int, y: str):
        logger.error(f"{x} {y}")

    magic = [156236, "Asd#g"]
    unexpected = f"{magic[0]} {magic[1]}"
    method_to_silence(*magic)
    assert unexpected not in ape_caplog.head



import pytest
from pydantic import BaseModel

from ape.api.address import Address, BaseAddress
from ape.types.address import AddressType


@pytest.fixture
def custom_address(zero_address):
    class MyAddress(BaseAddress):
        @property
        def address(self) -> AddressType:
            return zero_address

    return MyAddress()


def test_eq(zero_address, custom_address):
    assert custom_address == zero_address
    assert zero_address == custom_address


def test_contains_in_list(zero_address, custom_address):
    assert zero_address in [custom_address]
    assert custom_address in [zero_address]


def test_address_type_in_model(zero_address):
    class CustomModel(BaseModel):
        address: AddressType

    model = CustomModel(address=zero_address)
    assert model.address == zero_address


def test_balance(zero_address):
    address = Address(zero_address)
    actual = address.balance
    expected = 0
    assert actual == expected
    # Also show can compare directly to currency-str.
    assert actual == "0 ETH"
